"use strict";

const Socket = require("net").Socket;
const tls = require("tls");
const parseListUnix = require("./parseListUnix");

const fs = require("fs");
const path = require("path");
const promisify = require("util").promisify;
const fsReadDir = promisify(fs.readdir);
const fsMkDir = promisify(fs.mkdir);
const fsStat = promisify(fs.stat);

/**
 * Minimal requirements for an FTP client.
 */
class Client {  
    
    /**
     * Create a client instance.
     * @param {number} timeoutMillis Timeout to apply to control and data connections. Use 0 for no timeout.
     */
    constructor(timeoutMillis = 0) {
        // A timeout can be applied to the control connection.
        this._timeoutMillis = timeoutMillis;
        // The current task to be resolved or rejected.
        this._task = undefined;
        // A function that handles incoming messages and resolves or rejects a task.
        this._handler = undefined;
        // Options for TLS connections.
        this.tlsOptions = {};
        // The client can log every outgoing and incoming message.
        this.verbose = false;
        // The control connection to the FTP server.
        this.socket = new Socket();
        // The data connection to the FTP server.
        this.dataSocket = undefined;
    }

    /**
     * Closes control and data sockets.
     */
    close() {
        this.log("Closing sockets.");
        this._closeSocket(this._socket);
        this._closeSocket(this._dataSocket);
    }

    get socket() {
        return this._socket;
    }

    set socket(socket) {
        if (this._socket) {
            // Don't close the existing control socket automatically.
            // The setter might have been called to upgrade an existing connection.
            this._socket.removeAllListeners();
        }
        this._socket = this._setupSocket(socket);
        if (this._socket) {
            this._socket.setKeepAlive(true);
            this._socket.on("data", data => {
                const message = data.toString().trim();
                const code = parseInt(message.substr(0, 3), 10);
                this.log(`< ${message}`);
                this._respond({ code, message });
            });
        }
    }

    get dataSocket() {
        return this._dataSocket;
    }

    set dataSocket(socket) {
        this._closeSocket(this._dataSocket);
        this._dataSocket = this._setupSocket(socket);
    }

    /**
     * Returns true if TLS is enabled for the control socket.
     * @returns {boolean}
     */
    get hasTLS() {
        return this._socket && this._socket.encrypted === true;
    }

    /**
     * Send an FTP command.
     * @param {string} command
     */
    send(command) {
        // Don't log passwords.
        const message = command.startsWith("PASS") ? "> PASS ###" : `> ${command}`;
        this.log(message);
        this._socket.write(command + "\r\n");
    }

    /**
     * Send an FTP command and handle any response until the new task is resolved. This returns a Promise that
     * will hold whatever the handler passed on when resolving/rejecting its task.
     * @param {string} command
     * @param {HandlerCallback} handler
     * @returns {Promise<any>}
     */
    handle(command, handler) {
        return new Promise((resolvePromise, rejectPromise) => {
            this._handler = handler;
            this._task = {
                // When resolving or rejecting we also want the handler
                // to no longer receive any responses or errors.
                resolve: (...args) => {
                    this._handler = undefined;
                    resolvePromise(...args);
                },
                reject: (...args) => {
                    this._handler = undefined;
                    rejectPromise(...args);
                }
            };
            if (command !== undefined) {
                this.send(command);
            }
        });
    }

    /**
     * Logs message if client is verbose
     * @param {string} message 
     */
    log(message) {
        if (this.verbose) {
            console.log(message);
        }
    }

    _respond(payload) {
        if (this._handler) {
            this._handler(payload, this._task);
        }        
    }

    _setupSocket(socket) {
        if (socket) {
            // All sockets share the same timeout.
            socket.setTimeout(this._timeoutMillis);
            // Reroute any events to the single communication channel with the currently responsible handler. 
            // In case of an error, the following will happen:
            // 1. The current handler should handle the error by at least rejecting the associated task.
            // 2. This rejection will then reject the Promise associated with the task.
            // 3. This rejected promise will then lead to an exception in the user's application code.
            socket.once("error", error => this._respond({ error })); // An error will automatically close a socket.
            // Report timeouts as errors.
            socket.once("timeout", () => {
                socket.destroy(); // A timeout does not automatically close a socket.
                this._respond({ error: "Timeout" });
            });
        }
        return socket;
    }

    _closeSocket(socket) {
        if (socket) {
            socket.destroy();
        }
    }
}

// Public API
module.exports = {
    // Basic API
    Client,
    connect,
    send,
    useTLS,
    enterPassiveMode,
    list,
    download,
    upload,
    // Convenience
    login,
    useDefaultSettings,
    cd,
    pwd,
    uploadDir,
    downloadDir,
    removeDir,
    ensureDir,
    // Useful for custom extensions
    parseIPV4PasvResponse,
    upgradeSocket
};

/**
 * @typedef {Object} PositiveResponse
 * @property {number} code  The FTP return code parsed from the FTP return message.
 * @property {string} message  The whole unparsed FTP return message.
 */

/**
 * @typedef {Object} NegativeResponse
 * @property {Object|string} error  The error description.
 * 
 * Negative responses are usually thrown as exceptions, not returned as values.
 */

/**
 * Connect to an FTP server.
 * 
 * @param {Client} client
 * @param {string} host
 * @param {number} port
 * @return {Promise<PositiveResponse>}
 */
function connect(client, host, port) {
    client.socket.connect(port, host);
    return client.handle(undefined, (res, task) => {
        if (res.code === 220) {
            task.resolve(res);
        }
        else {
            task.reject(res);
        }
    });
}

/**
 * Send an FTP command. If successful it will return a response object that contains
 * the return code as well as the whole message.
 * 
 * @param {Client} client
 * @param {string} command
 * @param {boolean} ignoreError
 * @return {Promise<PositiveResponse>}
 */
function send(client, command, ignoreErrorCodes = false) {
    return client.handle(command, (res, task) => {
        const success = res.code >= 200 && res.code < 400;
        if (success || (res.code && ignoreErrorCodes)) {
            task.resolve(res);
        }
        else {
            task.reject(res);
        }
    });
}

/**
 * Upgrade the current socket connection to TLS.
 * 
 * @param {Client} client
 * @param {Object} options Same options as in `tls.connect(options)`
 * @return {Promise<PositiveResponse>}
 */
function useTLS(client, options) {
    return client.handle("AUTH TLS", (res, task) => {
        if (res.code === 200 || res.code === 234) {
            upgradeSocket(client.socket, options).then(tlsSocket => {
                client.log("Control socket is using " + tlsSocket.getProtocol());
                client.socket = tlsSocket; // TLS socket is the control socket from now on
                client.tlsOptions = options; // Keep the TLS options for later data connections that should use the same options.
                task.resolve(res);
            }).catch(err => task.reject(err));
        }
        else {
            task.reject(res);
        }  
    });
}

/**
 * Upgrade a socket connection.
 * 
 * @param {Socket} socket 
 * @param {Object} options Same options as in `tls.connect(options)`
 * @returns {Promise<TLSSocket>}
 */
function upgradeSocket(socket, options) {
    return new Promise((resolve, reject) => {
        options = Object.assign({}, options, { 
            socket // Establish the secure connection using the existing socket connection.
        }); 
        const tlsSocket = tls.connect(options, () => {
            const expectCertificate = options.rejectUnauthorized !== false;
            if (expectCertificate && !tlsSocket.authorized) {
                reject(tlsSocket.authorizationError);
            }
            else {
                resolve(tlsSocket);
            }
        });                
    });
}

/**
 * Prepare a data socket using passive mode.
 * 
 * @param {Client} client
 * @returns {Promise<PositiveResponse>}
 */
function enterPassiveMode(client, parsePasvResponse = parseIPV4PasvResponse) {
    return client.handle("PASV", (res, task) => {
        if (res.code === 227) {
            const target = parsePasvResponse(res.message);
            if (!target) {
                task.reject("Can't parse PASV response", res.message);
                return;
            }
            let socket = new Socket();
            socket.once("error", err => {
                task.reject("Can't open data connection in passive mode: " + err.message);
            });
            socket.connect(target.port, target.host, () => {
                if (client.hasTLS) {
                    // Upgrade to TLS, reuse TLS session of control socket.
                    const options = Object.assign({}, client.tlsOptions, { 
                        socket, 
                        session: client.socket.getSession()
                    });
                    socket = tls.connect(options);
                    client.log("Data socket uses " + socket.getProtocol());
                }
                socket.removeAllListeners();
                client.dataSocket = socket;
                task.resolve(res);    
            });
        }
        else {
            task.reject(res);
        }
    });   
}

/**
 * Parse a PASV response message.
 * 
 * @param {string} message
 * @returns {{host: string, port: number}}
 */
function parseIPV4PasvResponse(message) {
    // From something like "227 Entering Passive Mode (192,168,3,200,10,229)",
    // extract host and port.
    const groups = message.match(/([-\d]+,[-\d]+,[-\d]+,[-\d]+),([-\d]+),([-\d]+)/);
    if (!groups || groups.length !== 4) {
        return undefined;
    }
    return {
        host: groups[1].replace(/,/g, "."),
        port: (parseInt(groups[2], 10) & 255) * 256 + (parseInt(groups[3], 10) & 255)
    };
}

/**
 * List files and folders of current directory.`
 * 
 * @param {Client} client
 * @param {(rawList: string) => FileInfo[]} parseList
 * @return {FileInfo[]>}
 */
function list(client, parseList = parseListUnix) {
    // Some FTP servers transmit the list data and then confirm on the
    // control socket that the transfer is complete, others do it the
    // other way around. We'll need to make sure that we wait until
    // both the data and the confirmation have arrived.
    let ctrlDone = false;
    let rawList = "";
    let parsedList = undefined;
    return client.handle("LIST", (res, task) => {
        if (res.code === 150) { // Ready to download
            client.dataSocket.on("data", data => {
                rawList += data.toString();
            });
            client.dataSocket.once("end", () => {
                client.dataSocket = undefined;
                client.log(rawList);
                parsedList = parseList(rawList);
                if (ctrlDone) {
                    task.resolve(parsedList);
                }
            });
        }
        else if (res.code === 226) { // Transfer complete
            ctrlDone = true;
            if (parsedList) {
                task.resolve(parsedList);
            }
        }
        else if (res.code >= 400 || res.error) {
            client.dataSocket = undefined;
            task.reject(res);
        }
    }); 
}

/**
 * Upload stream data as a file. For example:
 * 
 * `upload(client, fs.createReadStream(localFilePath), remoteFilename)`
 * 
 * @param {Client} client 
 * @param {stream.Readable} readableStream 
 * @param {string} remoteFilename 
 * @returns {Promise<PositiveResponse>}
 */
function upload(client, readableStream, remoteFilename) {
    const command = "STOR " + remoteFilename;
    return client.handle(command, (res, task) => {
        if (res.code === 150) { // Ready to upload
            // If all data has been flushed, close the socket to signal
            // the FTP server that the transfer is complete.
            client.dataSocket.once("finish", () => client.dataSocket = undefined);
            readableStream.pipe(client.dataSocket);
        }
        else if (res.code === 226) { // Transfer complete
            task.resolve(res);
        }
        else if (res.code >= 400 || res.error) {
            client.dataSocket = undefined;
            task.reject(res);
        }
    });
}

/**
 * Download a remote file as a stream. For example:
 * 
 * `download(client, fs.createWriteStream(localFilePath), remoteFilename)`
 * 
 * @param {Client} client 
 * @param {stream.Writable} writableStream 
 * @param {string} remoteFilename 
 * @param {number} startAt 
 * @returns {Promise<PositiveResponse>}
 */
function download(client, writableStream, remoteFilename, startAt = 0) {
    const command = startAt > 0 ? `REST ${startAt}` : `RETR ${remoteFilename}`;
    return client.handle(command, (res, task) => {
        if (res.code === 150) { // Ready to download
            client.dataSocket.pipe(writableStream);
        }
        else if (res.code === 350) { // Restarting at startAt.
            client.send("RETR " + remoteFilename);
        }
        else if (res.code === 226) { // Transfer complete
            client.dataSocket = undefined;
            task.resolve(res);
        }
        else if (res.code >= 400 || res.error) {
            client.dataSocket = undefined;
            task.reject(res);
        }
    });
}

/**
 * Login
 * 
 * @param {Client} client 
 * @param {string} user 
 * @param {string} password 
 * @returns {Promise<PositiveResponse>}
 */
async function login(client, user, password) {
    await send(client, "USER " + user);
    return await send(client, "PASS " + password);
}

/**
 * Set some default settings.
 * 
 * @param {Client} client 
 */
async function useDefaultSettings(client) {
    await send(client, "TYPE I"); // Binary mode
    await send(client, "STRU F"); // Use file structure
    if (client.hasTLS) {
        await send(client, "PBSZ 0", true); // Set to 0 for TLS
        await send(client, "PROT P", true); // Protect channel (also for data connections)
    }
}

/**
 * Change the working directory.
 * 
 * @param {Client} client 
 * @param {string} remotePath
 * @returns {Promise<PositiveResponse>}
 */
function cd(client, remotePath) {
    return send(client, "CWD " + remotePath);
}

/**
 * Get the current working directory.
 * 
 * @param {Client} client 
 * @returns {Promise<string>}
 */
async function pwd(client) {
    const res = await send(client, "PWD");
    // The directory is part of the return message, for example: 
    // 257 "/this/that" is current directory.
    return res.message.match(/"(.+)"/)[1];
}

/**
 * Remove a directory of the current working directory. This will remove all files and directories recursively.
 * 
 * @param {Client} client 
 * @param {string} remoteDirPath 
 */
async function removeDir(client, remoteDirPath) {
    await send(client, "CWD " + remoteDirPath);
    await clearDir(client);
    // Remove the directory itself if we're not already on root.
    const workingDir = await pwd(client);
    if (workingDir !== "/") {
        await send(client, "CDUP");
        await send(client, "RMD " + remoteDirPath);        
    }
}

/**
 * Clear all contents of the current working directory. This does not remove
 * the working directory itself.
 * 
 * @param {Client} client 
 */
async function clearDir(client) {
    await enterPassiveMode(client);
    for (const file of await list(client)) {
        if (file.isDirectory) {
            await cd(client, file.name);
            await clearDir(client);
            await send(client, "CDUP");
            await send(client, "RMD " + file.name);
        }
        else {
            await send(client, "DELE " + file.name);
        }
    }
}

/**
 * Uploads a local directory to the current working directory.
 * 
 * @param {Client} client 
 * @param {string} localDirPath  A local path, e.g. "foo/bar" or "../test"
 * @param {string?} remoteDirName  (Optional) The name of the remote directory. If undefined, directory contents will be uploaded to the working directory.
 */
async function uploadDir(client, localDirPath, remoteDirName = undefined) {
    // If a remote directory name has been provided, create it and cd into it.
    if (remoteDirName !== undefined) {
        if (remoteDirName.indexOf("/") !== -1) {
            throw new Error(`Path provided '${remoteDirName}' instead of single directory name.`);
        }
        await send(client, "MKD " + remoteDirName);
        await cd(client, remoteDirName);
    }
    await uploadDirContents(client, localDirPath);
    // The working directory should stay the same after this operation.
    if (remoteDirName !== undefined) {
        await send(client, "CDUP");
    }
}

/**
 * Upload all contents of a local directory to the current working directory.
 * 
 * @param {Client} client 
 * @param {string} localDirPath 
 */
async function uploadDirContents(client, localDirPath) {
    const files = await fsReadDir(localDirPath);
    for (const file of files) {
        const fullPath = path.join(localDirPath, file);
        const stats = await fsStat(fullPath);
        if (stats.isFile()) {
            await enterPassiveMode(client);
            await upload(client, fs.createReadStream(fullPath), file);
        }
        else if (stats.isDirectory()) {
            await send(client, "MKD " + file);
            await cd(client, file);
            await uploadDirContents(client, fullPath);
            await send(client, "CDUP"); 
        }
    }
}

/**
 * Download all contents of the current working directory to a local directory.
 * 
 * @param {string} localDirPath 
 */
async function downloadDir(client, localDirPath) {
    await ensureLocalDirectory(localDirPath);
    await enterPassiveMode(client);
    for (const file of await list(client)) {
        const localPath = path.join(localDirPath, file.name);
        if (file.isDirectory) {
            await cd(client, file.name);
            await downloadDir(client, localPath);
            await send(client, "CDUP");
        }
        else {
            await enterPassiveMode(client);
            const writable = fs.createWriteStream(localPath);
            await download(client, writable, file.name);
        }
    }
}

async function ensureLocalDirectory(path) {
    try {
        await fsStat(path);
    }
    catch(err) {
        await fsMkDir(path);
    }    
}

/**
 * Make sure a given remote path exists, creating all directories as necessary. This
 * function also changes the current working directory to the given path.
 * 
 * @param {Client} client 
 * @param {string} remoteDirPath 
 */
async function ensureDir(client, remoteDirPath) {
    const names = remoteDirPath.split("/").filter(name => name !== "");
    // If the remoteDirPath was absolute go to root directory.
    if (remoteDirPath.startsWith("/")) {
        await cd(client, "/");
    }
    for (const name of names) {
        // Check first if the directory exists. Just calling MKD might
        // result in a permission error for intermediate directories.
        const res = await send(client, "CWD " + name, true);
        if (res.code >= 400) {
            await send(client, "MKD " + name);
            await cd(client, name);    
        }
    }
}
